import unittest
import configparser
import datetime
from pathlib import Path

from metaindex.configuration import Configuration, CONF_DEFAULTS
from metaindex.query import tokenize, Query
from metaindex.cache import Cache, MemoryCache, CacheBase
from metaindex.shared import CacheEntry


class TestTokenizer(unittest.TestCase):
    """Test suite for Query's tokenization function"""

    def test_simple_words(self):
        """Test the tokenization of terms"""
        text = '\\(there\\) once "was a"  \\"lady (in riga )'
        expected = ['(there)', 'once', 'was a', '\"lady', '(in', 'riga', ')']
        self.assertEqual(expected, tokenize(text))


class TestQueryParser(unittest.TestCase):
    """Test suite for Query.parse"""

    def test_simple(self):
        """Test for simple splitting into terms"""
        text = "look, albatross"
        query = Query.parse(text)

        self.assertEqual(2, len(query.root.terms))

    def test_keyvalue(self):
        """Basic parsing of key:value terms"""
        text = "title:albatross"
        query = Query.parse(text)

        self.assertEqual(query.root.terms[0].key, "title")
        self.assertEqual(query.root.terms[0].regex, "albatross")


class MatchingQueryBaseClass(unittest.TestCase):
    """Test suite for the Query.matches function"""
    def setUp(self):
        baseconfig = configparser.ConfigParser(interpolation=None)
        baseconfig.read_dict(CONF_DEFAULTS)
        baseconfig.set('General', 'cache', ':memory:')
        now = datetime.datetime.now()
        self.cache = self.setup_cache(baseconfig)
        if self.cache is None:
            return

        # fill with testable entries here
        self.cache.insert(CacheEntry('a',
                            {'extra.comment': ['something'],
                             'rules.tags': ['file'],
                             'extra.tags': ['data'],
                             'rules.title': ['Not a picture at all!'],
                            }, now))
        self.cache.insert(CacheEntry('b',
                            {'mimetype': ['image/png'],
                             'extra.title': ['Test picture', 'Not actually there'],
                            }, now))
        self.cache.insert(CacheEntry('c',
                            {'opf.title': ['A book with a title'],
                             'extra.type': ['book'],
                             'mimetype': ['application/epub+zip'],
                            }, now))
        self.cache.insert(CacheEntry('d',
                            [('mimetype', 'text/plain'),
                             ('extra.tag', 'foo'),
                             ('extra.tag', 'bar'),
                             ('extra.title', 'Document')], now))

    def setup_cache(self, baseconfig):
        self.skipTest('Base class')
        return None

    def test_find_all(self):
        """Just find all entries with a blank query"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find(''))

        self.assertEqual(len(result), 4)

    def test_word(self):
        """Test for simple word queries"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('some'))

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].path, Path('a'))

    def test_find_no_word(self):
        """Test that nothing is found when searching for a word that appears nowhere"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('nada'))

        self.assertEqual(len(result), 0)

    def test_find_regex(self):
        """Check for a simple regex match"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('s.+e[tT]hing'))

        self.assertEqual(len(result), 1)

    def test_find_no_regex(self):
        """Check for failing regex matches"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('si.+e[tT]hing'))

        self.assertEqual(len(result), 0)

    def test_simple_keyvalue(self):
        """A simple key value test case"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('rules.tags:file'))

        self.assertEqual(len(result), 1)

    def test_regex_keyvalue(self):
        """The value may be a regex, so test for it, too"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('extra.tags:d..a'))

        self.assertEqual(len(result), 1)

    def test_key_exists(self):
        """Test for the key existence operator"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('mimetype?'))

        self.assertEqual(len(result), 3)

    def test_synonym(self):
        """Test that synonyms are considered when doing key/value matches"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('title:picture'))

        self.assertEqual(len(result), 2)
        self.assertEqual(set(entry.path for entry in result), {Path('a'), Path('b')})

    def test_synonym_exists(self):
        """Test the existence checks for synonyms"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('tag?'))

        self.assertEqual(len(result), 2)

    def test_query_mimetype(self):
        """Test for querying by mimetype"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('mimetype:image/'))

        self.assertEqual(len(result), 1)

    def test_not_mimetype(self):
        """Test the inverted test for a property"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('not:mimetype:image/'))

        self.assertEqual(len(result), 3)

    def test_has_no_key(self):
        """Test that a key is not present"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('not:opf.title?'))

        self.assertEqual(len(result), 3)

    def test_more_conditions(self):
        """Test an actual sequence of conditions"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('title? not:mimetype?'))

        self.assertEqual(len(result), 1)
        self.assertEqual({r.path for r in result},
                         {Path('a'),})

    def test_find_tags(self):
        """Test a search for multiple tags"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('tag:file tag:data'))

        self.assertEqual(len(result), 1)
        self.assertEqual({r.path for r in result},
                         {Path('a')})

    def test_find_tags2(self):
        """Test a search for multiple tags"""
        assert isinstance(self.cache, CacheBase)
        result = list(self.cache.find('mimetype:text tag:foo tag:bar'))

        self.assertEqual(len(result), 1)
        self.assertEqual({r.path for r in result},
                         {Path('d')})


class NoCacheBackend(CacheBase):
    """A fake cache backend providing the API but nothing else"""
    def __init__(self, _):
        self.is_started = True

    def find(self, *_):
        return []

    def get(self, *_):
        return []

    def rename(self, *_):
        pass

    def insert(self, *_):
        pass

    def refresh(self, *_):
        pass

    def last_modified(self):
        return datetime.datetime.min

    def keys(self, *_):
        return set()

    def start(self):
        pass

    def quit(self):
        pass


class TestableMemoryCache(MemoryCache):
    BACKEND_TYPE = NoCacheBackend


class TestMatchingQueryOnMemoryCache(MatchingQueryBaseClass):
    def setup_cache(self, baseconfig):
        cache = TestableMemoryCache(Configuration(baseconfig))
        cache.wait_for_reload()
        assert len(cache.entries_by_path) == 0
        return cache

class TestMatchingQueryOnSqlCache(MatchingQueryBaseClass):
    def setup_cache(self, baseconfig):
        cache = Cache(Configuration(baseconfig))
        return cache



"""Some SQL Accessors, helpers, and constants"""
import datetime
import re
import sqlite3
import pathlib
from collections import namedtuple

from . import shared
from .query import QueryVisitor, Term


SQL_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
MIN_DATE = "0001-01-01T00:00:00"


class SqlAccess:
    """Basic SQLite wrapper"""
    TYPE_DATETIME = 'd'
    TYPE_INT = 'i'
    TYPE_FLOAT = 'f'
    TYPE_STR = None

    CREATE_FILE_TABLE = f"""
      create table if not exists `files`
          (`id` integer primary key autoincrement,
           `path` text unique,
           `last_modified` varchar(19) default "{MIN_DATE}")
    """
    CREATE_META_TABLE = """
      create table if not exists `metadata`
          (`id` integer references `files`(`id`) on delete cascade,
           `key` varchar(50) not null,
           `type` varchar(1) default null,
           `value` text not null,
           unique (`id`, `key`, `value`) on conflict ignore)
    """
    SELECT_METADATA_BASE_QUERY = """
      select `path`, `last_modified`, `metadata`.*
      from `files` left outer join `metadata` on `files`.`id` = `metadata`.`id`
    """

    def __init__(self, uri):
        """Create a persistent SQL access.

        :param uri: The location of the database. For example a path location
                    or ``:memory:``."""
        self.db = sqlite3.connect(uri)
        self.db.row_factory = sqlite3.Row
        self.db.create_function('REGEXP', 2, regexp)

        with self.db:
            self.db.execute(type(self).CREATE_FILE_TABLE)
            self.db.execute(type(self).CREATE_META_TABLE)

    def files(self):
        """Get all paths of files in the database

        :return: List of all file paths cached in the database
        :rtype: ``list[pathlib.Path]``
        """
        with self.db:
            return [pathlib.Path(row['path'])
                    for row in self.db.execute('select `id`, `path` from `files`')]

    def last_modified(self):
        """Get the timestamp of the most recently modified file in the database

        :rtype: ``datetime.datetime``
        """
        with self.db:
            query = 'select max(last_modified) from files'
            for row in self.db.execute(query):
                return str_to_dt(row[0])

    def keys(self):
        """Returns a set of all known metadata keys."""
        with self.db:
            return {row['key']
                    for row in self.db.execute("select distinct `key` from `metadata`")} | \
                   shared.CacheEntry.AUTO_KEYS

    def find(self, query):
        """Find and return all entries that match the query

        :param query: the search query to run
        :type query: ``metaindex.query.Query``
        :rtype: ``list[shared.CacheEntry]``
        """
        visitor = SqlQueryVisitor()
        query.accept(visitor)
        sqlquery, args = visitor.sqlquery, visitor.args
        args = [list(a) if isinstance(a, (set, tuple)) else a
                for a in args]
        results = {}
        with self.db:
            ids = [row['id'] for row in self.db.execute(sqlquery, args)]
            if len(ids) == 0:
                return []

            placeholder = ", ".join(["?"]*len(ids))

            sqlquery = type(self).SELECT_METADATA_BASE_QUERY + \
                       f" where `files`.`id` in ({placeholder})"

            for row in self.db.execute(sqlquery, ids):
                id_ = row['id']
                if id_ not in results:
                    last_modified = str_to_dt(row['last_modified'])
                    results[id_] = shared.CacheEntry(pathlib.Path(row['path']),
                                                    last_modified=last_modified)
                key = row['key']
                if key is None:
                    continue
                value = self._translate_row(row)
                results[id_].add(key, value)
        return list(results.values())

    def get(self, paths):
        """Get the metadata of these paths

        :param paths: list of paths to query
        :rtype: ``list[shared.CacheEntry]``
        """
        results = {}
        with self.db:
            for path in paths:
                pathresults = {}
                path = str(path)
                query = type(self).SELECT_METADATA_BASE_QUERY + \
                        " where substr(`path`, 1, ?) = ?"

                for row in self.db.execute(query, (len(path), path)):
                    id_ = row['id']

                    # found an entry twice
                    if id_ in results:
                        continue

                    if id_ not in pathresults:
                        last_modified = str_to_dt(row['last_modified'])
                        pathresults[id_] = shared.CacheEntry(pathlib.Path(row['path']),
                                                             last_modified=last_modified)
                    key = row['key']
                    if key is None:
                        continue
                    value = self._translate_row(row)
                    pathresults[id_].add(key, value)

                results.update(pathresults)
        return list(results.values())

    def _translate_row(self, row):
        type_, value = row['type'], row['value']

        if type_ == type(self).TYPE_DATETIME:
            value = str_to_dt(value)
        elif type == type(self).TYPE_INT:
            value = int(value)
        elif type == type(self).TYPE_FLOAT:
            value = float(value)
        else:
            value = str(value)

        return value

    def insert(self, items):
        """Insert ``items`` in the database

        This will overwrite all existing entries (by path) in the database.

        :param items: A list of ``CacheEntry`` to insert
        :return: Number of inserted entries
        """
        if len(items) == 0:
            return 0

        with self.db:
            # flush existing entries
            ids = set()
            query = "select `id` from `files` where `path` = ?"
            for item in items:
                ids |= {row['id']
                        for row in self.db.execute(query, (str(item.path),))}
            if len(ids) > 0:
                placeholder = ", ".join(["?"]*len(ids))
                self.db.execute(f"delete from `metadata` where `id` in ({placeholder})",
                                list(ids))
                self.db.execute(f"delete from `files` where `id` in ({placeholder})",
                                list(ids))

            # add all items
            for item in items:
                query = "insert into `files`(`path`, `last_modified`) values(?, ?)"
                result = self.db.execute(query,
                                         (str(item.path), dt_to_str(item.last_modified)))
                row_id = result.lastrowid

                for key, value in item:
                    if key in shared.CacheEntry.AUTO_KEYS:
                        continue

                    query = "insert into `metadata` values (?, ?, ?, ?)"
                    rawvalue = value.raw_value

                    if isinstance(rawvalue, int):
                        type_ = type(self).TYPE_INT
                        rawvalue = str(rawvalue)
                    elif isinstance(value, float):
                        type_ = type(self).TYPE_FLOAT
                        rawvalue = str(rawvalue)
                    elif isinstance(value, datetime.datetime):
                        type_ = type(self).TYPE_DATETIME
                        rawvalue = dt_to_str(rawvalue)
                    else:
                        type_ = type(self).TYPE_STR
                        rawvalue = str(rawvalue)

                    self.db.execute(query,
                                    (row_id, key, type_, rawvalue))
            return len(items)

    def flush(self):
        """Flush the entire database"""
        with self.db:
            self.db.execute("delete from metadata")
            self.db.execute("delete from files")

    def purge(self, paths):
        """Remove all entries with these paths from the database

        If a path points to a directory instead of a file, all
        files in all subdirectories will be removed, too.

        :param files: List of paths
        """
        if len(paths) == 0:
            return

        with self.db:
            ids = set()
            # match all files that start with the path
            pathquery = "select `id` from `files` where substr(`path`, 1, ?) = ?"
            for path in paths:
                path = str(path)
                ids |= {row['id']
                        for row in self.db.execute(pathquery, (len(path), path))}

            id_set = ", ".join(["?"]*len(ids))
            ids = list(ids)
            self.db.execute(f"delete from metadata where id in ({id_set})", ids)
            self.db.execute(f"delete from files where id in ({id_set})", ids)

    def expire_metadata(self, paths):
        """Mark the metadata of these paths as expired

        Expiring does not remove the data from the database, only marks it as
        very, very old.

        If a paths is pointing to a directory instead of a file, all metadata
        of all items in all the subdirectories will be expired, too.

        If ``paths`` is ``None``, all metadata will be expired.

        :param paths: the paths to expire, ``None`` will expire all metadata
        """

        if paths is None:
            with self.db:
                self.db.execute("update `files` set `last_modified` = ?",
                                (MIN_DATE,))
            return

        with self.db:
            for path in paths:
                strpath = str(path)
                self.db.execute("update `files` set `last_modified` = ? "
                                "where substr(`path`, 1, ?) = ?",
                                (MIN_DATE, len(strpath), strpath))

    def rename_file(self, old_path, new_path):
        """Rename a file from ``old_path`` to ``new_path``

        Only affects the database, no files on the filesystem are renamed"""
        with self.db:
            self.db.execute("update `files` set `path` = ? where `path` = ?",
                            (str(new_path), str(old_path)))

    def rename_dir(self, old_path, new_path):
        """Rename a directory from ``old_path`` to ``new_path``

        Only affects the database, no directories on the filesystem are renamed"""
        stroldpath = str(old_path)

        with self.db:
            query = 'select `id`, `path` from `files` where substr(`path`, 1, ?) = ?'
            id_path = [(row[0], pathlib.Path(row[1]))
                       for row in self.db.execute(query, (len(stroldpath), stroldpath))]
            id_path = [(id_,
                        (new_path / path.relative_to(old_path)).resolve())
                        for id_, path in id_path]
            for id_, newpath in id_path:
                self.db.execute('update `files` set `path` = ? where `id` = ?',
                                (str(newpath), id_))


SqlTerm = namedtuple('SqlTerm',
                     ['expression', 'args', 'inverted', 'operator'])


class SqlQueryVisitor(QueryVisitor):
    OR = ' union '
    AND = ' intersect '
    OPERATOR_MAPPING = {Term.OR: ' union ',
                        Term.AND: ' intersect '}

    def __init__(self):
        self.sequence_stack = []
        self.sqlquery = ''
        self.args = []

    def on_sequence_start(self, _):
        self.sequence_stack.append([])

    def on_sequence_end(self, element):
        sequence = self.sequence_stack.pop(-1)

        expr = ''
        args = []

        if len(self.sqlquery) > 0:
            expr += self.OPERATOR_MAPPING[element.operator]

        for index, subterm in enumerate(sequence):
            subexpr = ''
            if index > 0:
                if subterm.inverted:
                    expr += ' except '
                else:
                    expr += self.OPERATOR_MAPPING[subterm.operator]
            elif subterm.inverted:
                expr += ' select distinct `metadata`.`id` from `metadata` except '
            expr += ' select distinct `metadata`.`id` from `metadata` where '
            subexpr += subterm.expression
            expr += subexpr
            args += subterm.args

        if len(expr) > 0:
            if len(self.sqlquery) > 0:
                self.sqlquery += self.OPERATOR_MAPPING[element.operator]
            self.sqlquery += expr
            self.args += args

        if len(self.sequence_stack) == 0 and len(self.sqlquery) == 0:
            self.sqlquery = 'select distinct `metadata`.`id` from `metadata`'

    def on_key_value(self, element):
        if element.regex not in regex_cache:
            regex_cache[element.regex] = re.compile(element.regex, re.IGNORECASE)
        keycheck = " = ?"
        keys = [element.key]

        if isinstance(element.key, (set, list)):
            keys = list(element.key)
            keycheck = " in (" + ", ".join(["?"]*len(keys)) + ")"

        expr = f"(`metadata`.`key` {keycheck} AND `metadata`.`value` REGEXP ?)"
        term = SqlTerm(expr,
                       keys + [element.regex],
                       element.inverted,
                       element.operator)
        self.sequence_stack[-1].append(term)

    def on_key_exists(self, element):
        keys = element.key
        if isinstance(keys, (tuple, set)):
            keys = list(keys)
        if not isinstance(keys, list):
            keys = [keys]

        keycheck = " in (" + ", ".join(["?"]*len(keys)) + ")"

        term = SqlTerm(f"`metadata`.`key` {keycheck}",
                       keys,
                       element.inverted,
                       element.operator)
        self.sequence_stack[-1].append(term)

    def on_regex(self, element):
        if element.word not in regex_cache:
            regex_cache[element.word] = re.compile(element.word, re.IGNORECASE)

        # match against any value and even against the path
        term = SqlTerm("lower(`metadata`.`value`) REGEXP ?",
                       [element.word],
                       element.inverted,
                       element.operator)
        self.sequence_stack[-1].append(term)


# This is a cache of compiled regular expressions for use
# in SQLite queries
regex_cache = {}


def regexp(expr, item):
    """Regular expression support for sqlite.

    This function will be executed when regular expression is called from
    within sqlite3 through REGEXP.
    """
    global regex_cache

    if not isinstance(item, str):
        return False

    if expr not in regex_cache:
        regex_cache[expr] = re.compile(expr)

    result = regex_cache[expr].search(item)
    matches = result is not None

    return matches


def str_to_dt(text):
    """Convert text to naive datetime, expecting the SQL_DATETIME_FORMAT

    Returns datetime.datetime.min in case of errors"""
    try:
        return datetime.datetime.strptime(text, SQL_DATETIME_FORMAT)
    except ValueError:
        return datetime.datetime.min


def dt_to_str(dt):
    """Convert a datetime to text representation for use in the database"""
    return dt.strftime(SQL_DATETIME_FORMAT)



"""abc notation file indexer"""
from metaindex import logger
from metaindex.indexer import IndexerBase, only_if_changed


class ABCNotationIndexer(IndexerBase):
    """Indexer for abc music notation files"""
    NAME = 'abcnotation'
    ACCEPT = ['text/vnd.abc', '.abc']
    PREFIX = ('abc',)

    @only_if_changed
    def run(self, path, metadata, _):
        logger.debug("[%s] processing %s", self.NAME, path.name)

        with open(path, 'rt', encoding="utf-8") as filehandle:
            for line in filehandle:
                if line.startswith('T:'):
                    metadata.add('abc.title', line[2:].strip())
                if line.startswith('M:'):
                    metadata.add('abc.meter', line[2:].strip())
                if line.startswith('C:'):
                    metadata.add('abc.composer', line[2:].strip())
                if line.startswith('Z:'):
                    metadata.add('abc.transcription', line[2:].strip())
                if line.startswith('G:'):
                    metadata.add('abc.group', line[2:].strip())
                if line.startswith('K:'):
                    metadata.add('abc.key', line[2:].strip())
                    break



import datetime
import pathlib
import os
import re
import queue
import threading
import time

from metaindex import indexer
from metaindex import configuration
from metaindex import sql
from metaindex import stores
from metaindex import shared
from metaindex import logger
from metaindex.query import Query, QueryVisitor, Term
from metaindex.shared import CacheEntry


class CacheBase:
    """The basic Cache API"""
    def __init__(self, config=None):
        self.config = config or configuration.load()
        assert isinstance(self.config, configuration.Configuration)

        self.ignore_tags = self.config.list(configuration.SECTION_GENERAL,
                                            configuration.CONFIG_IGNORE_TAGS,
                                            "")
        """List of tags to not cache"""

    def refresh(self, paths=None, recursive=True, processes=None):
        """(Re-)Index all items found in the given paths or all cached items
           if paths is ``None``.

        If any item in the list is a directory and recursive is True, all
        items inside that directory (including any all subdirectories) will
        be indexed, too.

        processes may be set to a number of processes that will be used in
        parallel to index the files. If left at None, there will be as many
        processes launched as CPUs are available.
        """
        raise NotImplementedError()

    def cleanup(self):
        """Find and remove all entries in the cache that refer to no longer existing files"""
        raise NotImplementedError()

    def clear(self):
        """Remove everything from the cache"""
        raise NotImplementedError()

    def forget(self, paths):
        """Remove all ``paths`` from the cache.

        :param paths: The list of paths to remove from cache."""
        raise NotImplementedError()

    def expire_metadata(self, paths):
        """Remove all metadata associated to these paths

        But keep the paths in the database.
        """
        raise NotImplementedError()

    def find(self, query):
        """Find all items matching this query.

        query may either be a human-written search term or a Query instance.

        Returns a list of (path to file, metadata, last_modified) tuples.
        """
        raise NotImplementedError()

    def get(self, paths):
        """Get metadata for all items of paths

        paths may also be a single path instead of a list of paths.

        If any element of paths is pointing to a directory and recursive is
        set to True, get will return the metadata for all elements inside
        that directory (and their subdirectories'), too.
        """
        raise NotImplementedError()

    def keys(self):
        """Get all metadata keys

        :rtype: ``list[str]``
        """
        raise NotImplementedError()

    def rename(self, path, new_path, is_dir=None):
        """Rename all entries in the database that are 'path' to 'new_path'

        If 'path' is pointing to a directory, all files in the subdirectories
        will be renamed correctly, too.
        That means that this operation can affect many rows and take some time.

        If you already renamed path to new_path, you have to provide the `is_dir`
        parameter to indicate whether the item you renamed is a directory or not!
        """
        raise NotImplementedError()

    def last_modified(self):
        """Return the date and time of the entry that most recently updated in the database.
        :rtype: ``datetime.datetime``
        """
        raise NotImplementedError()

    def insert(self, item):
        """Insert the CacheEntry ``entry`` into the cache.

        This operation will not modify the item in the filesystem nor update
        any other form of metadata persistency for the item.
        This function really only affects the cache.
        """
        raise NotImplementedError()

    def parse_extra_metadata(self, metafile):
        """Extract extra metadata from this file"""
        data = stores.get_for_collection(metafile)

        for filename in data.keys():
            data[filename][shared.IS_RECURSIVE] = data[filename][shared.IS_RECURSIVE] and \
                                                  self.config.recursive_extra_metadata

        return data


class Cache(CacheBase):
    CACHE_FILENAME = "index" + os.extsep + "db"

    def __init__(self, config=None):
        super().__init__(config)

        # create database connection
        location = self.config.get(configuration.SECTION_GENERAL,
                                   configuration.CONFIG_CACHE,
                                   configuration.CACHEPATH)
        if location != ':memory:':
            location = pathlib.Path(location).expanduser().resolve() / Cache.CACHE_FILENAME

            if not location.parent.exists():
                location.parent.mkdir(parents=True, exist_ok=True)

        self.database = sql.SqlAccess(location)

    def refresh(self, paths=None, recursive=True, processes=None):
        if paths is None:
            paths = self.database.files()

        elif isinstance(paths, (set, tuple)):
            paths = list(paths)

        elif not isinstance(paths, list):
            paths = [paths]

        if len(paths) == 0:
            return []

        files = self.config.find_indexable_files(paths, recursive)

        # cache of last_modified dates for files
        last_modified = {fn: shared.get_last_modified(fn) for fn in files}

        # obtain cached metadata
        last_cached = {entry.path: entry for entry in self.get(files)}

        indexer_result = indexer.index_files(files,
                                             self.config,
                                             processes,
                                             self.config.ocr,
                                             self.config.extract_fulltext,
                                             last_modified,
                                             last_cached)

        results = []
        for result in indexer_result:
            if not result.success and not self.config.index_unknown:
                continue
            result.info.last_modified = last_modified[result.filename]
            results.append(result.info)
        self.database.insert(results)
        return results

    def cleanup(self):
        self.database.purge([path
                             for path in self.database.files()
                             if not path.exists()])

    def clear(self):
        self.database.flush()

    def expire_metadata(self, paths):
        if paths is None:
            pass
        elif not isinstance(paths, list):
            paths = [paths]

        if isinstance(paths, list):
            paths = [pathlib.Path(path).resolve()
                     for path in paths]

        self.database.expire_metadata(paths)

    def rename(self, path, new_path, is_dir=None):
        if (is_dir is not None and is_dir) or path.is_dir():
            self.database.rename_dir(path, new_path)
        else:
            self.database.rename_file(path, new_path)

    def last_modified(self):
        return self.database.last_modified()

    def forget(self, paths):
        self.database.purge(paths)

    def find(self, query):
        if query is None or len(query) == 0:
            query = Query()

        elif isinstance(query, str):
            query = Query.parse(query, self.config.synonyms)

        elif not isinstance(query, Query):
            raise TypeError()

        return self.database.find(query)

    def get(self, paths):
        if not isinstance(paths, (list, set, tuple)):
            paths = [paths]

        paths = [pathlib.Path(path).resolve() for path in paths]

        return self.database.get(paths)

    def insert(self, item):
        self.database.insert([item])

    def keys(self):
        return self.database.keys()


class ThreadedCache(CacheBase):
    """Special version of Cache to be used in multi-threaded applications

    To use this, create an instance and execute ``start``.
    """
    GET = "get"
    FIND = "find"
    REFRESH = "refresh"
    INSERT = "insert"
    GET_KEYS = "get-keys"
    RENAME = "rename"
    FORGET = "forget"
    CLEAR = "clear"
    CLEANUP = "cleanup"
    LAST_MODIFIED = "last_modified"
    EXPIRE = "expire"

    BACKEND_TYPE = Cache

    def __init__(self, config):
        super().__init__(config)
        self.queue = queue.Queue()
        self._quit = False
        self.handler = threading.Thread(target=self.handler_loop)
        self.results = queue.Queue()
        self.single_call = threading.Lock()
        self.cache = None
        self._started = False

    @property
    def is_started(self):
        """Whether or not the thread has been started"""
        return self._started

    def assert_started(self):
        """Ensure that the worker thread has been started"""
        if not self._started:
            raise RuntimeError("Not started")

    def get(self, paths):
        self.assert_started()
        with self.single_call:
            self.queue.put((self.GET, paths))
            return self.results.get()

    def find(self, query):
        self.assert_started()
        with self.single_call:
            self.queue.put((self.FIND, query))
            return self.results.get()

    def refresh(self, paths=None, recursive=True, processes=None):
        self.assert_started()
        with self.single_call:
            self.queue.put((self.REFRESH, paths, recursive, processes))
            return self.results.get()

    def insert(self, item):
        self.assert_started()
        with self.single_call:
            self.queue.put((self.INSERT, item))
            return self.results.get()

    def rename(self, path, new_path, is_dir=None):
        self.assert_started()
        with self.single_call:
            self.queue.put((self.RENAME, path, new_path, is_dir))
            return self.results.get()

    def forget(self, paths):
        self.assert_started()
        with self.single_call:
            self.queue.put((self.FORGET, paths))
            return self.results.get()

    def clear(self):
        self.assert_started()
        with self.single_call:
            self.queue.put((self.CLEAR,))
            return self.results.get()

    def cleanup(self):
        self.assert_started()
        with self.single_call:
            self.queue.put((self.CLEANUP,))
            return self.results.get()

    def expire_metadata(self, paths):
        self.assert_started()
        with self.single_call:
            self.queue.put((self.EXPIRE, paths))
            return self.results.get()

    def keys(self):
        self.assert_started()
        with self.single_call:
            self.queue.put((self.GET_KEYS,))
            return self.results.get()

    def last_modified(self):
        self.assert_started()
        with self.single_call:
            self.queue.put((self.LAST_MODIFIED,))
            return self.results.get()

    def start(self):
        """Launch the cache thread"""
        self.handler.start()

    def quit(self):
        """End the cache thread"""
        self._quit = True
        if self._started:
            self._started = False
            self.queue.put("")
            self.handler.join()

    def handler_loop(self):
        """The handler running in the separate thread.

        You do not need to call this. Call ``start`` instead.
        """
        if self._started:
            return
        self.cache = type(self).BACKEND_TYPE(self.config)
        self._started = True
        while not self._quit:
            item = self.queue.get()
            if len(item) < 1:
                continue

            command = item[0]
            args = item[1:]
            result = None

            try:
                if command == self.GET:
                    result = self.cache.get(*args)
                elif command == self.FIND:
                    result = self.cache.find(*args)
                elif command == self.REFRESH:
                    result = self.cache.refresh(*args)
                elif command == self.INSERT:
                    self.cache.insert(*args)
                elif command == self.LAST_MODIFIED:
                    result = self.cache.last_modified()
                elif command == self.GET_KEYS:
                    result = self.cache.keys()
                elif command == self.FORGET:
                    self.cache.forget(*args)
                elif command == self.CLEAR:
                    self.cache.clear()
                elif command == self.CLEANUP:
                    self.cache.cleanup()
                elif command == self.EXPIRE:
                    self.cache.expire_metadata(*args)
            except Exception as exc:
                result = exc

            self.results.put(result)


class MemoryCache(CacheBase):
    """Version of the threaded cache that uses in-memory caching of the database

    Upon initialisation this cache will obtain all entries from the database
    and try to keep the data in memory up to date.
    """
    BACKEND_TYPE = ThreadedCache

    def __init__(self, config):
        super().__init__(config)
        # TODO - refactor this into a proper directory tree
        #        to make several accesses (mostly directory-based)
        #        much faster
        self.entries_by_path = {}
        self.all_keys = set()
        self.last_read = datetime.datetime.min

        self.tcache = type(self).BACKEND_TYPE(config)
        self.writing = threading.Lock()
        self.reloading = threading.Lock()
        self.is_initialized = False

    def is_busy(self):
        """Whether or not the cache is busy reading from the database"""
        return self.writing.locked()

    def wait_for_write(self):
        """Wait until any currently pending write operation is completed"""
        with self.writing:
            # this will wait until it could acquire the lock and return again at once
            pass

    def wait_for_reload(self):
        """Wait until any currently pending reload operation is completed"""
        with self.reloading:
            # this will wait until it could acquire the lock and return again at once
            pass

    def check_dirty(self):
        """Check whether or not the database is more recent than this cache

        This call will return immediately.

        This call may return False even if it doesn't know for sure that the cache is dirty.

        However, if it returns True, you can be sure that the cache is dirty (i.e. no longer valid).
        """
        if self.writing.locked() or self.tcache.single_call.locked():
            return False

        last_modified = self.tcache.last_modified()
        return self.last_read > last_modified

    def start(self):
        """Call this before attempting any queries."""
        self.tcache.start()
        self.invalidate()

    def insert(self, item):
        with self.writing:
            self._do_insert(item)

        # run the actual renaming in the background though
        threading.Thread(target=lambda:
                self._persist_inserts([item])).start()

    def bulk_insert(self, inserts):
        """Insert a whole set of files and their metadata

        inserts is a list of `CacheEntry`s, just like you would use for 'insert'.
        """
        with self.writing:
            for item in inserts:
                self._do_insert(item)

        threading.Thread(target=lambda: self._persist_inserts(inserts)).start()

    def _do_insert(self, item):
        # assumes you know what you are doing and self.writing is acquired
        item.ensure_last_modified()
        self.entries_by_path[item.path] = item
        self.all_keys |= set(self.entries_by_path[item.path].keys())

    def _persist_inserts(self, inserts):
        for item in inserts:
            self.tcache.insert(item)

    def forget(self, paths):
        if not isinstance(paths, (list, tuple, set)):
            paths = [paths]

        to_purge = []
        with self.writing:
            # ugh, this is soooo inefficient. see the todo in __init__
            # about refactoring this
            for path in paths:
                strpath = str(path)
                to_delete = [key for key in self.entries_by_path
                                 if str(key).startswith(strpath)]
                to_purge += to_delete
                for key in to_delete:
                    del self.entries_by_path[key]

        threading.Thread(target=lambda: self._forget_from_cache(to_purge)).start()

    def _forget_from_cache(self, to_delete):
        self.tcache.forget(to_delete)

    def clear(self):
        with self.writing:
            self.entries_by_path = {}
            self.all_keys = set()

            self.tcache.clear()

    def cleanup(self):
        with self.writing:
            to_forget = [path for path in self.entries_by_path
                         if not path.exists()]
            for path in to_forget:
                del self.entries_by_path[path]

        threading.Thread(target=lambda: self._forget_from_cache(to_forget)).start()

    def expire_metadata(self, paths):
        with self.writing:
            self.tcache.expire_metadata(paths)

    def last_modified(self):
        return max(v.last_modified for v in self.entries_by_path.values())

    def find(self, query):
        """Find all entries that match this query"""
        matcher = CacheEntryQueryVisitor()
        query = Query.parse(query, synonyms=self.config.synonyms)
        query.accept(matcher)

        return [e for e in self.entries_by_path.values()
                if matcher.matches(e)]

    def get(self, paths):
        """Get all entries for these paths (recursively)"""
        if not isinstance(paths, (list, tuple, set)):
            paths = [paths]

        return [entry
                for path, entry in self.entries_by_path.items()
                if any(path.is_relative_to(pathpattern)
                       for pathpattern in paths)]

    def bulk_rename(self, renames):
        """Run the rename operation for all renames

        See also below, `rename`.

        `renames` must be a tuple `(old_path, new_path)`.
        `old_path` can be a directory, moving all cached subentries
        accordingly, too.

        You may add a `is_dir` boolean value to prevent the lookup of whether
        or not `old_path` is a directory, i.e. `(old_path, new_path, is_dir)`.
        Use it if you *first* rename the path and then update the cache.
        """
        cache_renames = []
        if any(len(data) not in [2, 3] for data in renames):
            raise ValueError()

        for data in renames:
            if len(data) == 2:
                path, new_path = data
                is_dir = None
            elif len(data) == 3:
                path, new_path, is_dir = data
            else:
                raise ValueError()

            if not isinstance(new_path, pathlib.Path):
                new_path = pathlib.Path(new_path).resolve()
            if not isinstance(path, pathlib.Path):
                path = pathlib.Path(path).resolve()

            logger.debug("Rename %s to %s", path, new_path)
            if (is_dir is not None and is_dir) or path.is_dir():
                entries = [e for p, e in self.entries_by_path.items()
                           if p.is_relative_to(path)]
                for entry in entries:
                    del self.entries_by_path[entry.path]
                    assert path.parts == entry.path.parts[:len(path.parts)]
                    parts = new_path.parts + entry.path.parts[len(path.parts):]
                    new_entry_path = pathlib.Path(os.sep.join(parts)).resolve()
                    logger.debug(" ... renaming %s to %s", entry.path, new_entry_path)
                    new_entry = CacheEntry(new_entry_path,
                                           entry.metadata,
                                           entry.last_modified)
                    self.entries_by_path[new_entry_path] = new_entry

            elif path in self.entries_by_path:
                entry = self.entries_by_path[path]
                new_entry = CacheEntry(new_path,
                                       entry.metadata,
                                       entry.last_modified)
                del new_entry['filename']
                new_entry.add('filename', str(new_path.name))
                self.entries_by_path[new_path] = new_entry
                del self.entries_by_path[path]

            cache_renames.append((path, new_path))

        threading.Thread(target=lambda: self.do_rename(cache_renames)).start()

    def rename(self, path, new_path, is_dir=None):
        """Move all metadata entries for 'path' to 'new_path'.

        Convenience wrapper for bulk_rename.
        """
        self.bulk_rename([(path, new_path, is_dir)])

    def do_rename(self, renames):
        """A blocking rename function.

        Do not call this directly, but call 'rename' instead."""
        with self.writing:
            for path, new_path in renames:
                self.tcache.rename(path, new_path)

    def refresh(self, paths=None, recursive=True, processes=None):
        """(Re-)index these paths (recursively by default)"""
        if paths is None:
            paths = list(self.entries_by_path.keys())
        if not isinstance(paths, (list, set, tuple)):
            paths = [paths]

        with self.reloading:
            with self.writing:
                self.tcache.refresh(paths, recursive, processes)
                if recursive:
                    for path in set(self.entries_by_path.keys()):
                        if any(path.is_relative_to(pathpattern) for pathpattern in paths):
                            del self.entries_by_path[path]
                else:
                    for path in paths:
                        if path in self.entries_by_path:
                            del self.entries_by_path[path]

                for entry in self.tcache.get(paths):
                    self.entries_by_path[entry.path] = entry
                    self.all_keys |= entry.metadata.keys()

    def keys(self):
        """Returns a set of all known metadata keys."""
        return self.all_keys

    def invalidate(self):
        """Invalidate the cached entries and reload"""
        self.entries_by_path = {}
        self.all_keys = set()
        self.reload()

    def reload(self):
        """Reload the data from the cache"""
        threading.Thread(target=self.do_reload).start()

    def do_reload(self):
        """A blocking reload function.

        Do not call this directly, but call 'invalidate' or 'reload' instead.
        """
        entries = {}
        keys = set()
        with self.reloading:
            attempt = 0
            while not self.tcache.is_started and attempt < 10:
                time.sleep(0.1)
                attempt += 1
            if not self.tcache.is_started:
                return

            for entry in self.tcache.find(''):
                entries[entry.path] = entry
                keys |= entry.metadata.keys()

            with self.writing:
                self.entries_by_path = entries
                self.all_keys = keys
                self.last_read = datetime.datetime.now()
                self.is_initialized = True

    def quit(self):
        """End the cache thread"""
        with self.writing:
            self.tcache.quit()
            self.tcache = ThreadedCache(self.config)
            self.entries_by_path = {}
            self.all_keys = set()


class CacheEntryQueryVisitor(QueryVisitor):
    """A QueryVisitor implementation to match against CacheEntry instances"""
    class Matcher:
        def __init__(self, visitor, element):
            self.visitor = visitor
            self.inverted = element.inverted
            self.operator = element.operator

        def match(self, item):
            raise NotImplementedError()

    class RegexMatcher(Matcher):
        def __init__(self, visitor, element):
            super().__init__(visitor, element)
            self.word = element.word

        def match(self, item):
            regex = self.visitor.regex_cache[self.word]
            matches = any(regex.search(str(value)) is not None
                          for value in set(value for value in item))
            if self.inverted:
                return not matches
            return matches

    class KeyExistsMatcher(Matcher):
        def __init__(self, visitor, element):
            super().__init__(visitor, element)
            self.keys = element.key
            if not isinstance(self.keys, (list, set, tuple)):
                self.keys = {self.keys}

        def match(self, item):
            result = any(key in item for key in self.keys)
            if self.inverted:
                return not result
            return result

    class KeyValueMatcher(Matcher):
        def __init__(self, visitor, element):
            super().__init__(visitor, element)
            self.keys = element.key
            self.regex = element.regex
            if not isinstance(self.keys, (list, tuple, set)):
                self.keys = {self.keys}

        def match(self, item):
            matches = any(sql.regex_cache[self.regex].search(str(value)) is not None
                          for value in sum([item[key] for key in self.keys], start=[]))
            if self.inverted:
                return not matches
            return matches

    class SequenceMatcher(Matcher):
        def __init__(self, visitor, element):
            super().__init__(visitor, element)
            self.matchers = []

        def match(self, item):
            # if len(self.matchers) > 1:
                #     breakpoint()
            results = []
            conjuncts = list(m.match(item) for m in self.matchers
                             if m.operator == Term.AND)
            disjuncts = list(m.match(item) for m in self.matchers
                             if m.operator == Term.OR)
            if len(conjuncts) > 0:
                results.append(all(conjuncts))
            if len(disjuncts) > 0:
                results.append(any(disjuncts))
            if len(conjuncts) == 0 and len(disjuncts) == 0:
                return True

            if self.operator == Term.AND:
                return all(results)
            return any(results)

    regex_cache = {}

    def __init__(self):
        self.sequence_stack = []
        self.root = None

    def on_regex(self, element):
        if element.word not in type(self).regex_cache:
            type(self).regex_cache[element.word] = re.compile(element.word, re.IGNORECASE)
        self.sequence_stack[-1].matchers.append(type(self).RegexMatcher(self, element))

    def on_key_exists(self, element):
        self.sequence_stack[-1].matchers.append(type(self).KeyExistsMatcher(self, element))

    def on_key_value(self, element):
        if element.regex not in type(self).regex_cache:
            sql.regex_cache[element.regex] = re.compile(element.regex, re.IGNORECASE)
        self.sequence_stack[-1].matchers.append(type(self).KeyValueMatcher(self, element))

    def on_sequence_start(self, element):
        self.sequence_stack.append(type(self).SequenceMatcher(self, element))

    def on_sequence_end(self, _):
        sequence = self.sequence_stack.pop(-1)
        if len(self.sequence_stack) == 0:
            self.root = sequence
        else:
            self.sequence_stack[-1].matchers.append(sequence)

    def matches(self, entry):
        if self.root is None:
            return True
        return self.root.match(entry)


if __name__ == '__main__':
    logger.setup('DEBUG')
    cache = Cache()
    breakpoint()



"""Collection sidecar file indexers"""
from metaindex import logger
from metaindex import json
from metaindex import opf
from metaindex import configuration
from metaindex import shared
from metaindex.indexer import IndexerBase, Order

try:
    from metaindex import yaml
except ImportError:
    yaml = None


class CollectionSidecarIndexer(IndexerBase):
    """Abstract collection sidecar indexer"""
    PREFIX = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        logger.debug(f"Creating a new {self.NAME}")

        # cached metadata for collection metadata, pathlib.Path -> multidict
        self.collection_cache = {}
        self.cached_sidecar_files = set()

        self.collection_metadata = self.cache.config.list(
            configuration.SECTION_GENERAL,
            configuration.CONFIG_COLLECTION_METADATA,
            '')
        self.recursive_metadata = self.cache.config.bool(
            configuration.SECTION_GENERAL,
            configuration.CONFIG_RECURSIVE_EXTRA_METADATA,
            'yes')

    def run(self, path, metadata, last_cached):
        raise NotImplementedError()

    def find_collection_sidecars(self, path, suffix):
        """Walk upwards the directory structure and yield all existing collection metadata files"""
        sidecars = []
        pathptr = path
        if pathptr.is_file():
            pathptr = path.parent
        while True:
            for filename in self.collection_metadata:
                sidecar = pathptr / (filename + suffix)
                if sidecar.exists():
                    sidecars.append(sidecar)
            if not self.recursive_metadata or pathptr == pathptr.parent:
                break
            pathptr = pathptr.parent

        return sidecars

    def cache_collection_sidecars(self, path, store):
        """Fill the cache with the sidecar files' contents"""
        assert self.PREFIX is not None
        for sidecar in self.find_collection_sidecars(path, store.SUFFIX):
            if sidecar in self.cached_sidecar_files:
                continue
            self.cached_sidecar_files.add(sidecar)

            self.collection_cache[sidecar.parent] = shared.CacheEntry(sidecar.parent)
            for filepath, extra in store.get_for_collection(sidecar).items():
                if filepath not in self.collection_cache:
                    self.collection_cache[filepath] = shared.CacheEntry(filepath)
                self.collection_cache[filepath].update(extra)

    def get_collection_metadata(self, path, suffix):
        meta = shared.CacheEntry(path)

        for sidecar in reversed(self.find_collection_sidecars(path, suffix)):
            if sidecar.parent not in self.collection_cache:
                logger.info("A sidecar file has been added while the indexer was running. "
                            "That file is ignored until the next rerun.")
                continue
            extra = self.collection_cache[sidecar.parent]
            if sidecar.parent != path.parent and \
               not (self.recursive_metadata and
                    any(v for k, v in extra if k == shared.IS_RECURSIVE)):
                continue
            meta.update(extra)

        if path in self.collection_cache:
            meta.update(self.collection_cache[path])

        return meta

    def run_with_store(self, store, path, metadata, _):
        assert self.PREFIX is not None
        self.cache_collection_sidecars(path, store)

        extra = self.get_collection_metadata(path, store.SUFFIX)

        # check for direct sidecar file
        sidecar = path.parent / (path.stem + store.SUFFIX)
        if sidecar.is_file() and sidecar != path:
            extra.update(store.get(sidecar))

        for key, value in extra:
            if key != shared.IS_RECURSIVE:
                metadata.add(key, value)


class JsonSidecarIndexer(CollectionSidecarIndexer):
    """JSON collection sidecar indexer"""
    NAME = 'json-sidecar'
    ACCEPT = '*'
    PREFIX = ''
    ORDER = Order.LAST

    def run(self, path, metadata, last_cached):
        return self.run_with_store(json, path, metadata, last_cached)


class OpfSidecarIndexer(CollectionSidecarIndexer):
    """OPF collection sidecar indexer"""
    NAME = 'opf-sidecar'
    ACCEPT = '*'
    PREFIX = ''
    ORDER = Order.LAST

    def run(self, path, metadata, _):
        self.cache_collection_sidecars(path, opf)
        for key, value in self.get_collection_metadata(path, opf.SUFFIX):
            if key != shared.IS_RECURSIVE:
                metadata.add(key, value)


if yaml is not None:
    class YamlSidecarIndexer(CollectionSidecarIndexer):
        """YAML collection sidecar indexer"""
        NAME = 'yaml-sidecar'
        ACCEPT = '*'
        PREFIX = ''
        ORDER = Order.LAST

        def run(self, path, metadata, last_cached):
            return self.run_with_store(yaml, path, metadata, last_cached)



from .cache import MemoryCache
from .shared import CacheEntry
from .query import Query
from .configuration import Configuration
from .indexer import IndexerBase, index_files, IndexerResult
from .humanizer import humanize, register_humanizer



"""Test cases for SqlAccess"""
import datetime
import unittest
from pathlib import Path

from metaindex import CacheEntry
from metaindex import Query
from metaindex.sql import SqlAccess, MIN_DATE, str_to_dt


class TestBaseCases(unittest.TestCase):
    def setUp(self):
        self.db = SqlAccess(":memory:")
        self.now = datetime.datetime.now().replace(microsecond=0)
        self.earlier = self.now - datetime.timedelta(minutes=4)
        self.min_date = str_to_dt(MIN_DATE)

    def test_files(self):
        self.test_insert()

        files = self.db.files()

        self.assertEqual({Path('/usr/share/doc/root.md'),
                          Path('/tmp/a.txt'),
                          Path('/tmp/b.txt')},
                         set(files))

    def test_last_modified(self):
        self.test_insert()

        last_modified = self.db.last_modified()

        self.assertEqual(last_modified, self.earlier)

    def test_keys(self):
        self.test_insert()

        keys = self.db.keys()

        self.assertEqual({'extra.tag', 'extra.title',
                          'extra.author',
                          'filename', 'size', 'last_modified'},
                         keys)

    def test_find(self):
        self.test_insert()
        result = self.db.find(Query())

        self.assertEqual(len(result), 3)
        self.assertTrue(all(isinstance(e, CacheEntry) for e in result))

        for entry in result:
            if entry.path == Path('/tmp/a.txt'):
                self.assertIn('extra.tag', entry)
                self.assertEqual(entry['extra.tag'], ['test'])
            elif entry.path == Path('/tmp/b.txt'):
                self.assertIn('extra.title', entry)
                self.assertEqual(set(entry['extra.title']),
                                 set(['tree', 'parrot']))
            elif entry.path == Path('/usr/share/doc/root.md'):
                self.assertIn('size', entry)
            else:
                self.fail(f"{entry.path} should not be here")

    def test_get(self):
        self.test_insert()

        results = self.db.get([Path('/usr/share/doc/root.md'),
                               Path('/tmp/b.txt'),
                               Path('/no/such/file.txt')])
        self.assertEqual({Path('/usr/share/doc/root.md'),
                          Path('/tmp/b.txt')},
                         {entry.path for entry in results})

    def test_insert(self):
        result = self.db.insert([
                     CacheEntry(Path('/tmp/a.txt'),
                                [('extra.tag', 'test'),
                                 ('extra.title', 'nothing really')],
                                self.earlier),
                     CacheEntry(Path('/tmp/b.txt'),
                                [('extra.title', 'parrot'),
                                 ('extra.title', 'tree'),
                                 ('extra.author', 'Caithlynn')],
                                self.now - datetime.timedelta(hours=3)),
                     CacheEntry(Path('/usr/share/doc/root.md'),
                                [('extra.title', 'Description of file hierarchy concept'),
                                 ('size', 42)],
                                self.now - datetime.timedelta(days=14))])
        self.assertEqual(result, 3)

    def test_flush(self):
        self.test_insert()

        self.db.flush()

        self.assertEqual(len(self.db.find(Query())), 0)

    def test_purge(self):
        self.test_insert()

        self.db.purge([Path('/tmp/a.txt'), Path('/usr/share/doc/root.md')])

        result = self.db.find(Query())

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].path, Path('/tmp/b.txt'))

    def test_expire_metadata(self):
        self.test_insert()

        self.db.expire_metadata([Path('/usr/share/doc/root.md'),
                                 Path('/tmp/b.txt')])

        results = {entry.path: entry.last_modified
                   for entry in self.db.find(Query())}
        expected = {Path('/tmp/a.txt'): self.earlier,
                    Path('/usr/share/doc/root.md'): self.min_date,
                    Path('/tmp/b.txt'): self.min_date}

        self.assertEqual(results, expected)

    def test_rename_file(self):
        self.test_insert()

        self.db.rename_file(Path('/tmp/a.txt'), Path('/tmp/new.txt'))

        results = self.db.find(Query())

        self.assertEqual({Path('/usr/share/doc/root.md'),
                          Path('/tmp/b.txt'),
                          Path('/tmp/new.txt')},
                         {entry.path for entry in results})

    def test_rename_dir(self):
        self.test_insert()

        self.db.rename_dir(Path('/tmp'), Path('/var/other'))

        results = self.db.find(Query())

        self.assertEqual({Path('/usr/share/doc/root.md'),
                          Path('/var/other/b.txt'),
                          Path('/var/other/a.txt')},
                         {entry.path for entry in results})
